# Dans ce script on nettoie deux tables qu'on est allé chercher sur la base de donnée de l'OCI.
# La première correspond à tous les articles francophones de mars à juin pour 2017 à 2019
# La seconde correspond à tous les articles francophones de mars à juin pour 2017 à 2020
# L'objectif est d'enlever les stopwords, la ponctuation, les chiffres et stemmatiser.

##IMPORTANT
# -> On doit rouler ce script sur un noeud de calcul de compute canada car il nécessite de faire de la parallélisation.
# Le script a été roulé avec 4 coeurs de 256G de mémoire.



#Chargement des librairies
require(dplyr)
require(data.table)
require(tidyr)
require(stringr)
require(tm)
require(tidytext)
require(doParallel)
require(fuzzyjoin)

#Ces path réfèrent à des répertoires sur Beluga dans le compte de l'équipe
source_folder <- "A_COMPLETER"
data_folder <- "A_COMPLETER"

if (!dir.exists(source_folder))
{
  stop("Ce script a été conçu rouler sur Beluga de Compute Canada")
}

## Chargement des fonctions ##
source(paste0(source_folder, "FCT_util.R"))

## Variables globales ##
testing <- FALSE #Mettre à TRUE pour faire des tests sur un plus petit jeu de données


#On prépare la parallélisation
nb_coeur <- 4
registerDoParallel(cores=nb_coeur)  
cl <- makeCluster(nb_coeur, type="FORK")
registerDoParallel(cl)

stopwords_ <- c("plus", "a", "’", "afin", "si", "an", "ça", "va", "fait", "très", "tout", "avoir", "être", "selon",
                "comme", "lors", "déjà", "dont", "aussi", "chez", "deux", "où", "entre", "e", "h", "quand", "jusquà", "dêtre", "quils",
                "na", "dun", "sest", "the", "parce", "ans", "faire", "dit", "tous")


# Liste de médias généralistes
liste_media <- read.csv(paste0(data_folder,"liste_finale_medias.csv")) %>% data.table()
liste_media <- liste_media[dans_oci == TRUE]


# Chargement des tables à nettoyer. Elles proviennent d'une requête à la base de l'OCI
# Ces tables ont 5 colonnes : _id, media_url, original_article_url, scraped_on (date) et full_article_text
load(paste0(data_folder, "dt_articles_2020.RData"))
load(paste0(data_folder, "dt_articles_2017_2019.RData"))



if (testing)
{
  #On prend un échantillon de 1000 articles
  dt_articles_2020 <- dt_articles_2020[runif(1000, 1, nrow(dt_articles_2020))]
  dt_articles_2017_2019 <- dt_articles_2017_2019[runif(1000, 1, nrow(dt_articles_2017_2019))]
}

#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
#### Nettoyage pour usage général ####
#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

dt_articles_2017_2019$corpus <- 1
dt_articles_2020$corpus <- 2
dt_full_text <- rbind(dt_articles_2017_2019, dt_articles_2020)
rm(dt_articles_2017_2019); rm(dt_articles_2020)
setnames(dt_full_text, "_id", "doc_ID")
#D'abord on met la date dans le bon format

dt_full_text[, date := as.Date(as.POSIXct(scraped_on, origin="1970-01-01"))][,scraped_on := NULL]

#On garde seulement les articles de plus de 200 caractères
dt_full_text[,text_length := str_length(full_article_text)]
dt_full_text <- dt_full_text[text_length > 200][,text_length := NULL]


#On prend seulement les médias que l'on veut dans notre corpus
dt_full_text <- take_medias_in_list(dt_full_text, liste_media)
#Certaines sections de média (sections affaires, argent, etc.) ne sont pas retirés par notre fonction
other_media_to_take_off <- c("https://www.985fm.ca/sports",
  "https://www.tvanouvelles.ca/argent",
  "http://www.lapresse.ca/techno/",
  "https://www.lapresse.ca/affaires/",
  "http://affaires.lapresse.ca/dossiers/la-presse-affaires-magazine/",
  "http://affaires.lapresse.ca/",
  "https://www.journalsaint-francois.ca/the-gleaner-changed-hands-wednesday/",
  "http://www.estrieplus.com/section-0404040431353537-actualites.html")

dt_full_text <- dt_full_text[!media_url %in% other_media_to_take_off][,media_url := NULL]


# Les acronymes qui reviennent souvent dans le corpus (CHSLD, SRAS, OMS, PCU)
# On veut éviter que des parties de la forme longue se retrouve dans les bigrams.
dt_full_text[,full_article_text := str_replace_all(full_article_text, 
                                                             pattern = "[Cc]entre[s]?(\\s)(d['’]hébergement|hospitaliers)(\\s)(et )?(de|en)(\\s)soins(\\s)(de(\\s))?longue(\\s)durée", 
                                                             replacement = "CHSLD")]
dt_full_text[,full_article_text := str_replace_all(full_article_text, 
                                                             pattern = "[Pp]restation(\\s)canadienne(\\s)d[’']urgence", 
                                                             replacement = "PCU")]
dt_full_text[,full_article_text := str_replace_all(full_article_text, 
                                                             pattern = "[Ss]yndrome(\\s)respiratoire(\\s)aigu(\\s)sévère", 
                                                             replacement = "SRAS")]
dt_full_text[,full_article_text := str_replace_all(full_article_text, 
                                                             pattern = "[Oo]rganisation(\\s)[Mm]ondiale(\\s)de(\\s)la(\\s)[Ss]anté", 
                                                             replacement = "OMS")]


#On nettoie le plein texte
dt_full_text[, cleaned_full_text := clean_corpus_parallel(full_article_text, 
                                                                    cl, additional_stopwords = stopwords_)]
dt_full_text[, cleaned_full_text := sapply(cleaned_full_text, as.character)]


print("nettoyage plein texte terminé")


#On tag les articles qui parlent de covid ou de science
dt_full_text[(grepl("covid", cleaned_full_text, ignore.case = TRUE) | 
             grepl("virus", cleaned_full_text, ignore.case = TRUE) |
             grepl("sars-cov", cleaned_full_text, ignore.case = TRUE))
             & corpus == 2, #Ils ne peuvent être que dans le second corpus (2020)
             is_covid := TRUE][is.na(is_covid), is_covid := FALSE]

dt_full_text[grepl("(\\s)scien", cleaned_full_text, ignore.case = TRUE) |
             grepl("(\\s)chercheu", cleaned_full_text, ignore.case = TRUE),
             is_science := TRUE][is.na(is_science), is_science := FALSE]

# La dernière partie prend du temps, donc on sauvegarde ici
save(dt_full_text, file = paste0(data_folder, "dt_full_text.RData"))

#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
#### Duplicats d'articles ####
#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

#Dans cette section, on veut donner un identifiant unique aux articles.
#Certains articles sont republiés tel quel par des médias appartenant au même conglomérat.
#Notre algorithme prend pour acquis que des articles republiés auront le même titre.
#IMPORTANT : certaines entrées de la BD ont le même texte même s'ils ont des titres différents.
#Nous ne traitons pas ces cas ici.

## On veut identifier les articles qui sont des duplicats
dt_titles_duplicates <- dt_full_text[, .N, by = "title"][N > 1][order(-N)]

#la documentation de fuzzy_join_article_duplicates est présente dans FCT_util.R
articles_list <- foreach(i = 1:nrow(dt_titles_duplicates),
                         .packages=c('tidyr', 'dplyr', 'fuzzyjoin', 'data.table'),
                         .combine = rbind) %dopar% fuzzy_join_article_duplicates(dt_titles_duplicates[i]$title, dt_full_text)

#L'ordonnancement des ID assure que l'un des ID d'article se retrouvera uniquement dans la colonne ID1
#On prend le art_ID de cet article
articles_list[!ID1 %in% ID2, art_ID := ID1, by="ID1"]
#Ensuite on bind tous les ID1 et les ID2 (qui auront le même art_ID)
articles_list <- rbindlist(list(articles_list[!is.na(art_ID), list(ID = ID1, art_ID, score)], 
                                articles_list[!is.na(art_ID), list(ID = ID2, art_ID, score)])) %>% unique()

#Si la distance est de 0.05 (nombre de changement par rapport à la longueur du texte) on considère
#que c'est le même texte
articles_list <- articles_list[score < 0.05]
#On merge les art_ID sur notre table principal
dt_full_text <- merge(dt_full_text, articles_list, by.x = "doc_ID", by.y = "ID", all.x = TRUE)
#Si une publication n'a pas de document, il n'a pas de doublons, donc on lui assigne son art_ID
dt_full_text[is.na(art_ID), art_ID := doc_ID][,score := NULL]

save(dt_full_text, file = paste0(data_folder, "dt_full_text.RData"))

