#Ce fichier contient des fonctions qui ont été développées pour le projet COVID, 
#mais elle sont suffisamment générales pour être utilisées pour d'autres projets.
#Ce fichier contient les fonctions clean_corpus(), BigramTokenizer(), get_unigram(),
# find_most_significant_bigram(), take_off_unigram_in_bigram() et take_medias_in_list()



clean_corpus <- function(corpus, stem = TRUE, language = "fr", additional_stopwords = "", preserve_dashes = TRUE)
{
  #This function uses tm functions and cleans a corpus. It removes punctuation, numbers, put everything to lower case
  #gets rid of stopwords and also stems words if the 'stem' argument is set to true.
  # [in] : corpus, a list of document you want to clean
  # [in] : language, the language of the corpus
  # [in] : additional_stopwords, a list of stopwords
  # [in] : preserve_dashes, a boolean telling if we want to preserves dashes within words
  # [out] : tible_clean_corpus, a tible of the cleaned documents
  corpus <- str_replace_all(corpus, "«", "")
  corpus <- str_replace_all(corpus, "»", "")
  corpus <- str_replace_all(corpus, "•", "")
  tible_clean_corpus <- VCorpus(VectorSource(corpus)) #On prend un échantillon de notre corpus 
  tible_clean_corpus <- tm_map(tible_clean_corpus, removePunctuation, preserve_intra_word_dashes = preserve_dashes)
  tible_clean_corpus <- tm_map(tible_clean_corpus, removeNumbers)
  tible_clean_corpus <- tm_map(tible_clean_corpus, content_transformer(tolower))
  stopwords <- c(stopwords(language), additional_stopwords)
  tible_clean_corpus <- tm_map(tible_clean_corpus, removeWords, stopwords)
  if(stem)
  {
    tible_clean_corpus <- tm_map(tible_clean_corpus, stemDocument)
  }
  tible_clean_corpus
}

clean_corpus_parallel <- function(corpus, cl, stem = TRUE, language = "fr", additional_stopwords = "", 
                                  preserve_dashes = TRUE)
{
  #This function uses tm functions and cleans a corpus. It removes punctuation, numbers, put everything to lower case
  #gets rid of stopwords and also stems words if the 'stem' argument is set to true.
  # [in] : corpus, a list of document you want to clean
  # [in] : language, the language of the corpus
  # [in] : additional_stopwords, a list of stopwords
  # [in] : preserve_dashes, a boolean telling if we want to preserves dashes within words
  # [out] : tible_clean_corpus, a tible of the cleaned documents
  corpus <- str_replace_all(corpus, "«", "")
  corpus <- str_replace_all(corpus, "»", "")
  corpus <- str_replace_all(corpus, "•", "")
  tm_parLapply_engine(cl)
  tible_clean_corpus <- VCorpus(VectorSource(corpus)) #On prend un échantillon de notre corpus 
  tible_clean_corpus <- tm_parLapply(tible_clean_corpus, removePunctuation, preserve_intra_word_dashes = preserve_dashes)
  tible_clean_corpus <- tm_parLapply(tible_clean_corpus, removeNumbers)
  tible_clean_corpus <- tm_parLapply(tible_clean_corpus, content_transformer(tolower))
  stopwords <- c(stopwords(language), additional_stopwords)
  tible_clean_corpus <- tm_parLapply(tible_clean_corpus, removeWords, stopwords)
  if(stem)
  {
    tible_clean_corpus <- tm_parLapply(tible_clean_corpus, stemDocument)
  }
  tible_clean_corpus
}

BigramTokenizer <- function(document)
{
  #This function takes a document and returns every bigram in it
  # [in] document, a string representing one document
  # [out] a list of bigrams
  unlist(lapply(ngrams(NLP::words(document), 2), paste, collapse = " "), use.names = FALSE)
}

UnigramTokenizer <- function(document)
{
  #This function takes a document and returns every unigram in it
  # [in] document, a string representing one document
  # [out] a list of unigrams
  unlist(lapply(ngrams(str_split(document, pattern = " "), 1), paste, collapse = " "), use.names = FALSE)
}
get_unigrams <- function(i)
{
  #generating all the unigrams.
  # [pre] There needs to be a table named corpus in the Global Environment. Corpus has 2 columns. full_text with the full text, and
  #       corpus with the number of the corpus the full text is in.
  # [pre] There needs to be a table name imp_bigrams in the Global Environment. It contains 3 columns. bigram with the bigram
  #       word1 with the first word of the bigram and word2 with the second word.
  # [in] i, the index of the row of the document we want to extract all unigrams of
  # [out] doc, a data.table with one unigram per row, and the corresponding document in second column
  doc <- str_split(corpus[i]$full_text, pattern = " ")[[1]]
  doc <- data.table(unigram = doc)
  #Take off unigram that are in our important bigrams from text.
  doc <- take_off_unigram_in_bigram(doc, imp_bigrams)
  #adding the cluster number.
  doc$corpus <- corpus[i]$corpus
  #returning
  doc %>% data.table()
}

find_most_significant_bigram <- function(corpus, colname = "V1", bigram_probability_thrhld = 0.25, parallel = FALSE)
{
  #This function takes a corpus (list of documents) and returns the most probable bigrams in that corpus. Stop words     
  #need to be removed in order to have good   results. The probability of a bigram is a conditionnal probability. It      
  #answers the question of what is the probability of the second word given the probability of the first one. We found    
  #that a probability of 0.25 (every bigram that has a probability of 0.25 or higher) has good results. The corpus must   
  #be a full text that went through a stemming process.
  
  # [in] corpus, a data.table containing the list of document
  # [in] colname, the name of the column in which the the documents are (default in "V1")
  # [in] bigram_probability_thrhld, the threshold over which we keep the bigram (default is 0.25)
  # [in] parallel, a bool saying if you want to parallelize the treatment or not
  # [out] imp_bigram, a data.table with the most important bigrams. It contains 3 columns, word1, word2 and bigram.
  # [pre] if you want to parallelize, you need to registerParallel before launching the function
  if (!"V1" %in% colnames(corpus))
  {
    setnames(corpus, colname, "V1")
  }
  dt_bigram <- data.table()
  #getting every bigram in the corpus
  if (parallel)
  {
    dt_bigram <- foreach(i = 1:nrow(corpus), .packages=c('tidyr', 'dplyr', 'data.table', 'stringr'), .combine = rbind) %dopar% data.table(BigramTokenizer(corpus[i]$V1), document = i)
  }
  else
  {
    for (i in c(1:nrow(corpus)))
    {
      doc_bigram <- data.table(BigramTokenizer(corpus[i]$V1), document = i)
      dt_bigram <- rbind(dt_bigram, doc_bigram, fill = TRUE)
    }
  }
  
  
  #First we count the frequency of every bigram
  dt_bigram[,bigram_freq := .N, by = "V1"]
  dt_bigram[,bigram := V1]
  
  #We need to know what are the two words of the bigram
  dt_bigram <- dt_bigram %>% separate(V1, c("word1", "word2"), sep = " ") %>% unique()
  
  #Getting rid of bigram that appears only once, since they are obviously not the most significant
  dt_bigram <- dt_bigram[bigram_freq > 1]
  
  #To calculate the probability of a bigram, we need the frequency of every unigram
  dt_unigram <- data.table(word = words(corpus$V1))
  dt_unigram <- dt_unigram[,word_freq :=  .N, by = "word"] %>% unique()
  
  #Merging that frequency on the bigram table
  dt_bigram <- merge(dt_bigram, dt_unigram, by.x = "word1", by.y = "word", all.x = TRUE, all.y = FALSE)
  
  #calculting the bigram probability
  dt_bigram[,bigram_prob := bigram_freq/word_freq]
  setorder(dt_bigram, -bigram_prob, -bigram_freq)
  
  #Getting the bigrams with a probability over our threshold
  imp_bigrams <- dt_bigram[bigram_prob > bigram_probability_thrhld, list(word1, word2, bigram)] %>% unique()
  setnames(corpus, "V1", colname)
  imp_bigrams
}

take_off_unigram_in_bigram <- function(document, imp_bigrams)
{
  #This function takes a list of unigram and takes off the unigram that are part of bigrams present in
  #the imp_bigrams table. The unigrams have to be in the same order as they appear in the document.
  # [in] document, a data.table with a single column named 'unigram' (one word per row)
  # [in] imp_bigrams, a data.table containing the important bigrams
  # [out] document_new without the unigrams that are in important bigrams
  
  #Recreating the full text (needed for BigramTokenizer)
  doc <- paste0(document$unigram, collapse = " ")
  #Getting all the bigrams
  bigrams <- data.table(BigramTokenizer(doc))
  #Marking which row has a bigram that is in our imp_bigrams list
  bigrams[V1 %in% imp_bigrams$bigram ,is_in_imp_bigrams := TRUE][is.na(is_in_imp_bigrams), is_in_imp_bigrams := FALSE]
  #We want to know which row has important bigrams because we need to get rid of the first word of the next bigram
  index_which_is_imp_bigram <- bigrams$is_in_imp_bigram %>% which
  #If index_which_is_imp_bigram is empty, there is nothing left to do
  if (length(index_which_is_imp_bigram) == 0)
  {
    bigrams <- bigrams %>% separate(V1, c("word1", "word2"), sep = " ")
    document_new <- rbind(bigrams[,list(unigram = word1)], bigrams[nrow(bigrams), list(unigram = word2)])
    
  }else{
    # If the index is the last one, we don't want to take off the first word of the next row cause it doesn't exists
    if (max(index_which_is_imp_bigram) == (nrow(bigrams)))
    {
      index_which_is_imp_bigram <-index_which_is_imp_bigram[-length(index_which_is_imp_bigram)]
    }
    #Separating the bigram in its two words
    bigrams <- bigrams %>% separate(V1, c("word1", "word2"), sep = " ")
    #Since the second word of a bigram is the first of the next one, we have to get rid of that one too
    bigrams[(index_which_is_imp_bigram + 1), word1 := ""]
    bigrams <- bigrams[is_in_imp_bigrams != TRUE & word1 != ""]
    document_new <- rbind(bigrams[,list(unigram = word1)], bigrams[nrow(bigrams), list(unigram = word2)])
  }
  document_new
}

take_medias_in_list <- function(query_result, media_list)
{
  #This function takes a list of media, with their url_root (named media_url_short) and filters a table in order to       keep entries that are from that media.
  # [in] query result, a data.table containing the results of a query to mongoDB with a column named 'media_url'
  # [in] media_list, a table containing a list of medias, with a column named 'media_url_short'
  # [out] query_result filtered
  
  if (!"media_url" %in% colnames(query_result))
  {
    stop('The query_result table has to have a column named "media_url"') 
  }
  if (!"media_url_short" %in% colnames(media_list))
  {
    stop('The media_list table has to have a column named "media_url_short"') 
  }
  
  #Extracting root_url
  query_result[,root_url:= str_split(str_split(media_url,"://")[[1]][2],pattern = "/")[[1]][1],by=media_url]
  #Returning the filtered table
  setkey(query_result, root_url)
  setkey(media_list, media_url_short)
  query_result[root_url %in% unique(media_list$media_url_short)]
}

fuzzy_join_article_duplicates <- function(title_, dt_articles)
{
  #Cette fonction trouve les doublons parmis des articles ayant le même titre
  #[in] title_ le titre d'un article qui revient plus d'une fois dans notre corpus
  #[in] dt_articles une table avec au moins une colone title, une colone full_article_text
  #     et une colonne doc_ID
  #[out] Une table avec tous les couples de doc_ID et un score
  #[note]  0.05 est le treshold au dessous duquel un paire d'article peuvent être considéré
  #        comme similaires.
  
  
  #getting all publications with that title
  dt_article <- dt_articles[title == title_]
  dt_article[,text_length := str_length(full_article_text)]
  #Over about 20000 charaters, the fuzzy join function overload our memory
  memory_limit_nb_character <- 10000
  dt_article[text_length > memory_limit_nb_character,  `:=` 
             (full_article_text = substr(full_article_text, 1, memory_limit_nb_character), text_length = memory_limit_nb_character)]
  #We did tests and concluded if the levenshtein distance made edits under 0.05 * (nb_character_text1 + nb_character_text2) / 2. It must be
  #the same article
  distance_proportion <- 0.05
  #We need a max distance in the stringdist_inner_join function so it don't take forever.
  max_distance <- max(dt_article$text_length) * distance_proportion
  fuzzy_join_distances <- stringdist_inner_join(dt_article, dt_article, by="full_article_text",
                                                distance_col = "distance",
                                                method = "osa",
                                                max_dist = max_distance) %>% data.table()
  
  #renaming
  setnames(fuzzy_join_distances,c("doc_ID.x", "doc_ID.y"), c("ID1", "ID2"))
  #Getting rid of distances between the same article
  fuzzy_join_distances <- fuzzy_join_distances[ID1 != ID2, list(ID1, ID2, distance, text_length.x, text_length.y)]
  #This will make sure we can have two or more different articles in a set of publications with the same title
  fuzzy_join_distances <- fuzzy_join_distances[, `:=`(ID1 = c(ID1,ID2) %>% sort() %>%  first(), ID2 = c(ID1,ID2) %>% sort() %>% last())
                                               , by = .(ID1,ID2)] %>% unique()
  
  #Returning the table with the score
  fuzzy_join_distances[,score := distance / (text_length.x + text_length.y / 2)]
}
